/*  error.cpp
    functions for plaintext error messages

    todo: -

    30.05.99: tk, initial version.

    Copyright (c)2000 by Thomas Kindler, thomas.kindler@gmx.de

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version. Read the
    full License at http://www.gnu.org/copyleft for more details.
*/

// include files ---------------
//
#include <string.h>
#include <stdio.h>
#include "ide/mpplayer.hpp"
#include "ide/atapi.hpp"
#include "ide/atapierror.hpp"


typedef struct ATAPI_SENSECODE {
  int     asc;
  int     ascq;
  char   *msg;
} ATAPI_SENSECODE;


static char *atapi_sensekeys[] = {
  "no sense",         "recovered error",
  "not ready",        "medium error",
  "hardware error",   "illegal request",
  "unit attention",   "data protect",
  "sense key 0x08",   "sense key 0x09",
  "sense key 0x0a",   "aborted command",
  "sense key 0x0c",   "sense key 0x0d",
  "miscompare",       "sense key 0x0f"
};


static ATAPI_SENSECODE  atapi_sensecodes[] =  {
   {  0x00, 0x00, "no additional sense information"  },
   {  0x00, 0x11, "play operation in progress"  },
   {  0x00, 0x12, "play operation paused"  },
   {  0x00, 0x13, "play operation successfully completed"  },
   {  0x00, 0x14, "play operation stopped due to error"  },
   {  0x00, 0x15, "no current audio status to return"  },
   {  0x01, 0x00, "mechanical positioning or changer error"  },
   {  0x02, 0x00, "no seek complete"  },
   {  0x04, 0x00, "logical drive not ready - cause not reportable"  },
   {  0x04, 0x01, "logical drive not ready - in progress of becoming ready"  },
   {  0x04, 0x02, "logical drive not ready - initializing command required"  },
   {  0x04, 0x03, "logical drive not ready - manual intervention required"  },
   {  0x05, 0x01, "media load - eject failed"  },
   {  0x06, 0x00, "no reference position found"  },
   {  0x09, 0x00, "track following error"  },
   {  0x09, 0x01, "tracking servo failure"  },
   {  0x09, 0x02, "focus servo failure"  },
   {  0x09, 0x03, "spindle servo failure"  },
   {  0x11, 0x00, "unrecovered read error"  },
   {  0x11, 0x06, "circ unrecovered error"  },
   {  0x15, 0x00, "random positioning error"  },
   {  0x15, 0x01, "mechanical positioning or changer error"  },
   {  0x15, 0x02, "positioning error detected by read of medium"  },
   {  0x17, 0x00, "recovered data with no error correction applied"  },
   {  0x17, 0x01, "recovered data with retries"  },
   {  0x17, 0x02, "recovered data with positive head offset"  },
   {  0x17, 0x03, "recovered data with negative head offset"  },
   {  0x17, 0x04, "recovered data with retries and/or circ applied"  },
   {  0x17, 0x05, "recovered data using previous sector id"  },
   {  0x18, 0x00, "recovered data with error correction applied"  },
   {  0x18, 0x01, "recovered data with error correction & retries applied"  },
   {  0x18, 0x02, "recovered data - the data was auto-reallocated"  },
   {  0x18, 0x03, "recovered data with circ"  },
   {  0x18, 0x04, "recovered data with l-ec"  },
   {  0x1a, 0x00, "parameter list length error"  },
   {  0x20, 0x00, "invalid command operation code"  },
   {  0x21, 0x00, "logical block address out of range"  },
   {  0x24, 0x00, "invalid field in command packet"  },
   {  0x26, 0x00, "invalid field in parameter list"  },
   {  0x26, 0x01, "parameter not supported"  },
   {  0x26, 0x02, "parameter value invalid"  },
   {  0x28, 0x00, "not ready to ready transition, medium may have changed"  },
   {  0x29, 0x00, "power on, reset or bus device reset occurred"  },
   {  0x2a, 0x00, "parameters changed"  },
   {  0x2a, 0x01, "mode parameters changed"  },
   {  0x30, 0x00, "incompatible medium installed"  },
   {  0x30, 0x01, "cannot read medium - unknown format"  },
   {  0x30, 0x02, "cannot read medium - incompatible format"  },
   {  0x39, 0x00, "saving parameters not supported"  },
   {  0x3a, 0x00, "medium not present"  },
   {  0x3f, 0x00, "atapi cd-rom drive operating conditions have changed"  },
   {  0x3f, 0x01, "microcode has been changed"  },

   // 0x40, 0xnn, "diagnostic failure on component nn (80h-ffh)"

   {  0x44, 0x00, "internal atapi cd-rom drive failure"  },
   {  0x4e, 0x00, "overlapped commands attempted"  },
   {  0x53, 0x00, "media load or eject failed"  },
   {  0x53, 0x02, "medium removal prevented"  },
   {  0x57, 0x00, "unable to recover table of contents"  },
   {  0x5a, 0x00, "operator request or state change input (unspecified)"  },
   {  0x5a, 0x01, "operator medium removal request"  },
   {  0x63, 0x00, "end of user area encountered on this track"  },
   {  0x64, 0x00, "illegal mode for this track"  },
   {  0xb9, 0x00, "play operation oborted"  },
   {  0xbf, 0x00, "loss of streaming"  },
   {  0, 0, NULL  },
};


char *ATAPI_GetErrorMsg( ATAPI_COMMAND *cmd )
{
  static char   msg[256];

  int senseKey = (cmd->status & 0xf0) >> 4;

  if (senseKey)  {
    sprintf( msg, "%s:", atapi_sensekeys[(cmd->status  & 0xf0) >> 4] );

    ATAPI_SENSECODE  *p = atapi_sensecodes;
    while (p->msg) {
      if ((p->asc == cmd->asc) && (p->ascq == cmd->ascq))
        break;
      p++;
    }

    if (p->msg)  strcat( msg, p->msg );
    else         strcat( msg, "unkown sense data" );
  }  else  sprintf( msg, "ok." );

  return msg;
}
