/*  iso9660.cpp
    ISO/IEC 9660:1999 CD-ROM filesystem reader

    todo: -

    06.06.99: tk, first real working version; reads standard iso and joliet
                  volumes, data recursion is used for building the directory
                  tree (fine for stack-space constrained systems), and we can
                  limit the maximum size taken up by the filelist.
    03.06.99: tk, joliet support added, code cleaned up a bit.
    02.06.99: tk, initial version.

    Copyright (c)2000 by Thomas Kindler, thomas.kindler@gmx.de

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version. Read the
    full License at http://www.gnu.org/copyleft for more details.
*/

// include files ---------------
//
#include <string.h>
#include "ide/mpplayer.hpp"
#include "ide/miscide.hpp"
#include "ide/cdrom.hpp"
#include "ide/iso9660.hpp"
#include "ide/isostructs.hpp"


static ERROR  ISO_ReadIsoPVD( ISO_PVD *pvd, ISO_VOLUME *vol )
{
  IsoToCString( vol->volumeID, pvd->volumeID, 32, false );

  vol->root.lba    = pvd->root.extentLocation;
  vol->root.len    = pvd->root.dataLength;
  vol->root.flags |= ISO_DIR_NODE;

  vol->volumeSize    = pvd->volumeSpaceSize;
  vol->isJoliet      = false;
  vol->isInitialized = true;
  return ERR_OK;
}


static ERROR  ISO_ReadJolietSVD( ISO_SVD *svd, ISO_VOLUME *vol )
{
  // check for joliet escape sequence
  if (   strncmp(svd->escSequences, JOLIET_UCS2_1, 3)
      && strncmp(svd->escSequences, JOLIET_UCS2_2, 3)
      && strncmp(svd->escSequences, JOLIET_UCS2_3, 3))
    return ERR_NOJOLIETSVD;

  // note: the volume_id field is 32 bytes = 16 unicode chars.
  IsoToCString( vol->volumeID, svd->volumeID, 16, false );

  vol->root.lba    = svd->root.extentLocation;
  vol->root.len    = svd->root.dataLength;
  vol->root.flags |= ISO_DIR_NODE;

  vol->volumeSize    = svd->volumeSpaceSize;
  vol->isJoliet      = true;
  vol->isInitialized = true;
  return ERR_OK;
}


ERROR  ISO_ReadVolumeDescriptor( ISO_VOLUME *vol )
{
  ERROR           error;
  ISO_VD_UNKNOWN  vdesc;

  memset( vol, 0, sizeof(ISO_VOLUME) );

  for (ulong lba = 16; ;lba++)  {
    error = CD_ReadBlocks( &vdesc, lba, 1 );
    if (error)  return  error;

    // check for ISO signature
    if ( strncmp(vdesc.header.stdID, ISO_STANDARD_ID, 5) )
      return  ERR_NOISO9660;

    switch (vdesc.header.type)  {
      case ISO_VDESC_PVD:
        // read primary volume descriptor.
        ISO_ReadIsoPVD( (ISO_PVD*)&vdesc, vol );
        break;

      case ISO_VDESC_SVD:
        // read secondary volume descriptor, while overwriting
        // the PVD. this relies on the assumption, that the last
        // descriptor is always the most recent, or in this case
        // the better one (i.e. joliet).
        ISO_ReadJolietSVD( (ISO_SVD*)&vdesc, vol );
        break;

      case ISO_VDESC_END:
        // return error if we haven't found a valid PVD/SVD yet,
        // success otherwise.
        return  vol->isInitialized ? ERR_OK : ERR_NOISO9660;
    }
  }
}


static far char bigbadbuf[ISO_FILE_LIST_SIZE];


ERROR  ISO_AddFileNode( ISO_VOLUME *vol, int fnLen )
{
  ISO_FILE_NODE  *node;

  fnLen += sizeof(ISO_FILE_NODE);

  // alloc filelist memory
  //
  if (!vol->fileList)  {
    vol->fileList = (ISO_FILE_NODE*)bigbadbuf;  // malloc(ISO_FILE_LIST_SIZE);
    if (!vol->fileList)  return  ERR_OUTOFMEMORY;

    memset( vol->fileList, 0, ISO_FILE_LIST_SIZE );

    vol->root.next = vol->fileList;
  }

  if (vol->fileListLength+fnLen > ISO_FILE_LIST_SIZE)
    return  ERR_TOOMANYFILES;

  node = (ISO_FILE_NODE*)( (char*)vol->fileList + vol->fileListLength );
  vol->fileListLength += fnLen;

  if (vol->lastFile)
    vol->lastFile->next = node;

  vol->lastFile = node;

  return ERR_OK;
}


ERROR  ISO_ReadDirectory( ISO_VOLUME     *vol,
                          ISO_FILE_NODE  *parentDirNode )
{
  ERROR           error;
  char            block[2048];
  char            fn[256];
  ulong           lba = parentDirNode->lba;
  ulong           len = parentDirNode->len;

  ISO_FILE_NODE   *currentDirNode;

  while (len > 0)  {
    error = CD_ReadBlocks( block, lba++, 1 );
    if (error)  return error;

    ISO_FRECORD *frec = (ISO_FRECORD*)block;
    while(frec->recordLength)  {

      // check for special filenames first
      //
      if (frec->fileID[0] == '\0' && frec->fileIDLength==1)
        strcpy( fn, "." );
      else if (frec->fileID[0] == '\1' && frec->fileIDLength==1)
        strcpy( fn, ".." );
      else {
        if (vol->isJoliet)
          IsoToCString( fn, frec->fileID, frec->fileIDLength/2, true );
        else
          IsoToCString( fn, frec->fileID, frec->fileIDLength, false );
      }

      error = ISO_AddFileNode( vol, strlen(fn) );
      if (error)  return  error;

      strcpy( vol->lastFile->filename, fn );
      vol->lastFile->lba   = frec->extentLocation;
      vol->lastFile->len   = frec->dataLength;

      if (!strcmp(vol->lastFile->filename, ".")) {
        // special link to current directory
        currentDirNode        = vol->lastFile;
        vol->lastFile->dir    = currentDirNode;
        vol->lastFile->flags |= ISO_CURRENT_DIR_NODE;
      } else if (!strcmp(vol->lastFile->filename, "..")) {
        // special link to parent directory
        parentDirNode->dir    = vol->lastFile;
        vol->lastFile->dir    = parentDirNode;
        vol->lastFile->flags |= ISO_PARENT_DIR_NODE;
      } else {
        // normal directory node
        vol->lastFile->dir    = currentDirNode;
        vol->lastFile->flags  = (frec->flags & 2) ? ISO_DIR_NODE : 0;
      }

      // advance to next node
      frec = (ISO_FRECORD*)( (char*)frec + frec->recordLength );
    }

    len -= 2048;
  }

  return ERR_OK;
}


ERROR  ISO_BuildDirectoryTree( ISO_VOLUME  *vol )
{
  ERROR  error;
  ISO_FILE_NODE  *node;

  vol->dirNum    = 0;
  vol->fileNum   = 0;
  vol->fileBytes = 0;

  for (node = &vol->root; node; node = node->next)  {
    if (node->flags & ISO_DIR_NODE)  {
      if (!(node->flags & (ISO_CURRENT_DIR_NODE | ISO_PARENT_DIR_NODE))) {
        error = ISO_ReadDirectory( vol, node );
        if (error)  return error;

        vol->dirNum++;
      }
    } else {
      vol->fileNum++;
      vol->fileBytes += node->len;
    }
  }
  return ERR_OK;
}


ERROR  ISO_InitVolume( ISO_VOLUME *vol )
{
  ERROR  error;

  // read volume descriptors
  //
  error = ISO_ReadVolumeDescriptor( vol );
  if (error)  return error;

  error = ISO_BuildDirectoryTree( vol );
//  if (error)  return error;


  return error;
}
